'''
To parse and get the path of all parts
'''

import os
from liaar.lib import exception, setting


def get_resources_abs_path(app_name):
    '''
    Validate and return the resource directory
    '''
    abs_resources_dir = os.path.abspath(os.path.join(get_app_abs_path(app_name),
                                                     setting.RESOURCES_DIRECTORY))

    if os.path.isdir(abs_resources_dir):
        return abs_resources_dir
    else:
        exception.handle('Invalid resources directory: %s' % abs_resources_dir)


def get_resource_abs_path(app_name, resource_name):
    resource_abs_path = '%s/%s' % (get_resources_abs_path(app_name), resource_name)

    if os.path.isdir(resource_abs_path):
        return resource_abs_path
    else:
        exception.handle('Resource path does not exist: %s' % resource_name)


def get_app_abs_path(app_name):
    '''
    Get application absolute path
    '''
    app_abs_path = os.path.abspath(os.path.join(setting.APPS_DIRECTORY,
                                                app_name))

    if os.path.isdir(app_abs_path):
        return app_abs_path
    else:
        exception.handle('Invalid application directory: %s' % app_abs_path)


def get_app_file(filename):
    '''
    Get filename + application file extension
    '''
    return filename + '.' + setting.APP_FILE_EXTENSION


def get_method_filename(app_name, resource_name, method_name):
    method_abs_path = '%s/%s' % (get_resource_abs_path(app_name, resource_name), get_app_file(method_name))
    if os.path.isfile(method_abs_path):
        return method_abs_path
    else:
        exception.handle('Method file does not exist: %s' % method_name)


def get_app_setting_filename(app_name):
    '''
    Get application setting filename
    '''
    app_abs_path = get_app_abs_path(app_name)

    app_setting_filename = os.path.join(app_abs_path,
                                        get_app_file(setting.APP_SETTING_FILENAME))

    if os.path.isfile(app_setting_filename):
        return app_setting_filename
    else:
        exception.handle('The given application doesn\'t have setting file: %s' %
                        get_app_file(setting.APP_SETTING_FILENAME))



import argparse
from liaar.lib import app, path, parser


# to print the tool manual and other messages
def greeting():
    print """
 _      _
| |    (_)
| |     _  __ _  __ _ _ __
| |    | |/ _` |/ _` | '__|
| |____| | (_| | (_| | |
|______|_|\__,_|\__,_|_|
"""


def intro():
    # show greeting message
    greeting()

    # config arguments
    parser = argparse.ArgumentParser(description='Liaar creates a fake REST API using JSON files')

    return parser.parse_args()


def init_apps():
    '''
    To initialize and add all applications
    '''
    apps_list = app.get_apps()
    # add apps
    for app_name in apps_list:
        app_setting = app.get_app_setting(app_name)
        app.add_app(app_name, app_setting['version'], app_setting)

        # add resources
        resources_list = app.get_resources(app_name)
        for resource_name in resources_list:
            app.add_resource(app_name, app_setting['version'], resource_name)

            methods_list = app.get_methods(app_name, resource_name)

            # add methods
            for method_name in methods_list:
                method_filename = path.get_method_filename(app_name,
                                                           resource_name,
                                                           method_name)
                method_json = parser.load_and_parse(method_filename)
                app.add_method(app_name,
                               app_setting['version'],
                               resource_name,
                               method_name,
                               method_json)



'''
Handle twisted resources and routing
'''
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.internet import reactor
from liaar.lib import route, exception, logger, setting, request_handler
import json


class SiteResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        logger.info('Processing request for %s' % request.path)
        route_item = route.get(request.postpath)

        if route_item is None:
            exception.handle('No route found for: %s' % request.path)

        route_object, route_params = route_item

        # set route name
        route_name = route_object[1]

        if route_name == 'root':
            return 'root'
        elif route_name == 'app':
            return 'app: %s' % route_params
        elif route_name == 'resource':
            return 'resource'
        elif route_name == 'method':
            return request_handler.method(request, route_params)

def start():
    root = SiteResource()
    factory = Site(root)
    reactor.listenTCP(setting.LIAAR_PORT, factory)
    logger.info('Liaar is available on port %d' % setting.LIAAR_PORT)
    # run the twisted
    reactor.run()



'''
Simple loggger
'''
import datetime


log_levels = {
    2: 'info',
    1: 'warning',
    0: 'error'
}

def log(message, level):
    '''
    Log a message with given log level
    '''
    print '[%s] [%s] %s' % (datetime.datetime.today().isoformat(), log_levels[level], message)


def info(message):
    '''
    Log an info message
    '''
    #TODO: we should select the log level from `log_levels`
    log(message, 2)



'''
To manage twisted routes
'''
import re
from liaar.lib import setting


# keep routes list
_routes = []


def add(route, name):
    '''
    Add a new route to routes list
    '''
    _routes.append((route, name))


def init():
    add('/', 'root')

    add(setting.APP_URL_FORMAT, 'app')

    add(setting.APP_URL_FORMAT +
        setting.APP_RESOURCE_URL_FORMAT, 'resource')

    add(setting.APP_URL_FORMAT +
        setting.APP_RESOURCE_URL_FORMAT +
        setting.APP_METHOD_URL_FORMAT, 'method')


def get_all():
    '''
    Get routes list
    '''
    return _routes


def extract_params(route, route_regex, path):
    '''
    Extract values and parameters from given path and return a dictionary
    '''
    params = re.findall('{([a-zA-Z0-9_]+)}', route[0])
    values = re.findall(route_regex, path)
    extracted_params = {}
    i = 0
    for param in params:
        param_value = values[0][i]
        # trim `v` character from `version` parameter
        if param == 'version':
            param_value = values[0][i].lstrip('v')

        extracted_params[param] = param_value
        i = i + 1

    return extracted_params


def get(arr_path):
    '''
    Verify and return a route with given array of postpath from twisted
    '''
    # creating the full path of the request
    path = '/' + '/'.join(arr_path)

    for route in _routes:
        route_regex = re.sub('{[a-zA-Z0-9_]+}', '([a-zA-Z0-9_]+)', route[0])
        if re.match('^' + route_regex + '$', path):
            return route, extract_params(route, route_regex, path)
            break



'''
To retrieve and parse application setting and resources file
'''

from liaar.lib import path, parser, exception, logger, setting
import re
import os


# to hold all applications' configuration in memory
apps = {}


def add_app(app_name, version, setting):
    '''
    Add a new application to in-memory dictionary
    '''
    apps[app_name] = {}
    apps[app_name][version] = {}
    apps[app_name][version]['resources'] = {}
    apps[app_name][version]['setting'] = setting
    # add log
    logger.info('Adding application `%s`' % app_name)


def add_resource(app_name, version, resource_name):
    '''
    Add a new resource to in-memory dictionary
    '''
    if not apps.has_key(app_name):
        exception.handle('Undefined `app_name`, call `add_app` before `add_resource`')

    if not apps[app_name].has_key(version):
        exception.handle('Undefined app version')

    apps[app_name][version]['resources'][resource_name] = {}

    # add log
    logger.info('Adding resource `%s` to application `%s`' % (resource_name, app_name))


def add_method(app_name, version, resource_name, method_name, method_setting):
    '''
    Add a new method to in-memory dictionary
    '''
    if not apps.has_key(app_name):
        exception.handle('Undefined `app_name`, call `add_app` before `add_resource`')

    if not apps[app_name].has_key(version):
        exception.handle('Undefined app version')

    if not apps[app_name][version]['resources'].has_key(resource_name):
        exception.handle('Undefined resource name')

    apps[app_name][version]['resources'][resource_name][method_name] = method_setting

    # add log
    logger.info('Adding method `%s` to resource `%s`' % (method_name, resource_name))


def get_method(app_name, version, resource_name, method_name):
    '''
    Get method's setting
    '''
    try:
        return apps[app_name][version]['resources'][resource_name][method_name]
    except KeyError:
        exception.handle('No method found')


def get_apps():
    '''
    To get all available applications
    '''

    apps_list = []

    apps_path = os.path.abspath(setting.APPS_DIRECTORY)

    for app_item in os.listdir(apps_path):
        # check if it's a directory
        app_abs_path = os.path.join(apps_path, app_item)
        if os.path.isdir(app_abs_path):
            apps_list.append(app_item)

    return apps_list


def get_resources(app_name):
    '''
    Get the list of resource directories
    '''
    resources_list = []

    resources_path = path.get_resources_abs_path(app_name)

    # get all files and folders from resource directory
    for resource in os.listdir(resources_path):
        # check if it's a directory
        resource_abs_path = os.path.abspath(os.path.join(resources_path,
                                                         resource))
        if os.path.isdir(resource_abs_path):
            resources_list.append(resource)

    return resources_list


def get_methods(app_name, resource_name):
    '''
    Get all methods of given resource name
    '''
    methods_list = []

    resource_abs_path = path.get_resource_abs_path(app_name, resource_name)

    for method_name in os.listdir(resource_abs_path):
        name_re = re.match('^(.+)\.json$', method_name)
        if name_re:
            methods_list.append(name_re.group(1))

    return methods_list


def get_app_setting(app_name):
    '''
    Get application's setting
    '''
    parsed_file = parser.load_and_parse(path.get_app_setting_filename(app_name))
    # add application name to setting dictionary
    parsed_file['name'] = app_name
    return parsed_file



'''
Twisted request handlers
'''
from liaar.lib import app, parser
import json


def method(request, route_params):
    # set response content-type to json
    request.responseHeaders.addRawHeader(b"content-type", b"application/json")
    method_setting = app.get_method(route_params['app_name'],
                                    route_params['version'],
                                    route_params['resource_name'],
                                    route_params['method_name'])

    return json.dumps(parser.parse_param(method_setting))



'''
 Liaar v0.2.0
 By Afshin Mehrabani (@afshinmeh) - 2014
'''
from lib import setting, bootstrap, route, twisted_handler


def main():
    # bootstrap
    args = bootstrap.intro()

    route.add('/', 'root')
    route.add(setting.APP_URL_FORMAT, 'app')
    route.add(setting.APP_URL_FORMAT +
              setting.APP_RESOURCE_URL_FORMAT, 'resource')
    route.add(setting.APP_URL_FORMAT +
              setting.APP_RESOURCE_URL_FORMAT +
              setting.APP_METHOD_URL_FORMAT, 'method')

    bootstrap.init_apps()

    twisted_handler.start()


if __name__ == '__main__':
    main()



'''
Default settings for liaar
'''


# default liaar port number
LIAAR_PORT = 1234

# apps directory
APPS_DIRECTORY = './apps/'

# resources directory
RESOURCES_DIRECTORY = './resources/'

# resource files extension
APP_FILE_EXTENSION = 'json'

# app setting file name
APP_SETTING_FILENAME = 'setting'

# app default url format
APP_URL_FORMAT = '/{app_name}/{version}'

# app's resources default url format
APP_RESOURCE_URL_FORMAT = '/{resource_name}'

# app's methods default url format
APP_METHOD_URL_FORMAT = '/{method_name}'



'''
In order to manage and handle exceptions and errors
'''


def handle(message, exception_detail=None):
    '''
    Raises an exception
    '''
    if exception_detail is not None:
        raise Exception(message + ' - %s' % exception_detail)
    else:
        raise Exception(message)



'''
To produce URLs for resources, apps, etc.
'''

from liaar.lib import setting


def app_url(app_setting):
    '''
    Get the application's URL
    '''
    return setting.APP_URL_FORMAT.format('v' + app_setting['version'], app_setting['name'])


def resource_url(resource_name, app_setting):
    '''
    Get the resource's URL
    '''
    return app_url(app_setting) + setting.APP_RESOURCE_URL_FORMAT.format(resource_name)


def method_url(method_name, resource_name, app_setting):
    '''
    Get the method's URL
    '''
    return resource_url(resource_name, app_setting) + setting.APP_METHOD_URL_FORMAT.format(method_name)



'''
To parse json files inside applications (setting.json, resource etc.)
'''

import json
from faker import Factory
from liaar.lib import exception


def load_and_parse(filename):
    '''
    Load and parse JSON file
    '''
    with open(filename, 'r') as f:
        try:
            return json.load(f)
        except ValueError as detail:
            exception.handle('Failed to parse the JSON.', detail)


def parse_param(value):
    '''
    Parse resources JSON setting file parameters
    '''

    formatter_name = None
    is_list = False
    list_count = 10

    # if the value is a string (single value)
    if isinstance(value, basestring):
        formatter_name = value
    # or if it's a dictionary
    elif isinstance(value, dict):
        if 'liaar_formatter' in value:
            formatter_name = value['liaar_formatter']
        if 'liaar_type' in value:
            is_list = True
        if 'liaar_count' in value:
            list_count = value['liaar_count']

        nested_result = {}
        # now we should iterate over keys and call `parse_param` again
        if formatter_name is None:
            for value_item in value:
                nested_result[value_item] = parse_param(value[value_item])

            return nested_result
    else:
        exception.handle('Unknown resource property: %s' % value)

    if formatter_name is None:
        exception.handle('Formatter is empty: %s' % value)

    # ok let's execute the function
    try:
        faker = Factory.create()

        if not is_list:
            return faker.__getattribute__(formatter_name)()
        else:
            list_output = []
            for i in range(list_count):
                list_output.append(faker.__getattribute__(formatter_name)())

            return list_output
    except AttributeError:
        exception.handle('Can\'t parse resource\'s property: %s' % value)



from distutils.core import setup
import setuptools


setup(
    name='Liaar',
    version='0.2.0',
    packages=['liaar', 'liaar.lib'],
    license='LICENSE',
    url='https://github.com/usablica/liaar',
    description='Create fake REST API in a minute',
    author='Afshin Mehrabani',
    author_email='afshin.meh@gmail.com',
    install_requires=['Twisted', 'fake-factory'],
    entry_points={
        'console_scripts': [
            'liaar = liaar.liaar:main',
        ]
    }
)
